//
//  HelloWorldLayer.m
//  nukeControl
//
//  Created by Patrick Völcker on 22.06.11.
//  Copyright herontronic 2011. All rights reserved.
//


// Import the interfaces
#import "HelloWorldLayer.h"

#define decElectronSpeed 4
#define thereIsNoBadElectron 99

// HelloWorldLayer implementation
@implementation HelloWorldLayer
@synthesize electrons, electronPath, atoms, bg;


+(CCScene *) scene
{
    // 'scene' is an autorelease object.
    CCScene *scene = [CCScene node];
    
    // 'layer' is an autorelease object.
    HelloWorldLayer *layer = [HelloWorldLayer node];
    
    // add layer as a child to scene
    [scene addChild: layer];
    
    // return the scene
    return scene;
}

-(void) restartLevel {
    
    // alles loeschen
    [self removeAllChildrenWithCleanup:YES];
    [bg removeAllObjects];
    [atoms removeAllObjects];
    [electrons removeAllObjects];
    [electronPath removeAllObjects];
    
    // Hintergrundanimation neu erzeugen
    for (int p=0;p<20;p++) {
        // Sprite
        CCSprite *bgAtom = [CCSprite spriteWithFile:
                            (p%2==0?@"atom1.png":@"atom2.png")];
        bgAtom.position = ccp(fmod(arc4random(),screenWidth),
                              fmod(arc4random(),screenHeight*2));
        [self addChild:bgAtom];
        [bg addObject:bgAtom];
    }
    
    // Atome neu erzeugen (in 3 Schichten fuer Licht- und Farbeffekte)
    NSMutableArray *atoms1 = [[[NSMutableArray alloc] init] autorelease];
    for (int p=0;p<3;p++) {
        CCSprite *singleAtom1 = [CCSprite spriteWithFile:
                                 (p%2==0?@"atom2.png":@"atom1.png")];
        [self addChild:singleAtom1];
        [atoms1 addObject:singleAtom1];
    }
    // zweites Atom erzeugen
    NSMutableArray *atoms2 = [[[NSMutableArray alloc] init] autorelease];
    for (int p=0;p<3;p++) {
        CCSprite *singleAtom2 = [CCSprite spriteWithFile:
                                 (p%2==0?@"atom1.png":@"atom2.png")];
        [self addChild:singleAtom2];
        [atoms2 addObject:singleAtom2];
    }
    [atoms addObject:atoms1];
    [atoms addObject:atoms2];
    
    // Elektronen neu erzeugen
    for (int p=0;p<electronsTotal;p++) {
        // Sprite
        CCSprite *singleElectron = [CCSprite spriteWithFile:@"electron.png"];
        singleElectron.position = ccp(40+p*(400/(electronsTotal-1)),
                                      screenHeight/2-25+arc4random()%50);
        [self addChild:singleElectron z:1 tag:0];
        [electrons addObject:singleElectron];
        
        // jedes Elektron hat zu Beginn 2 Pfadpunkte:
        // 1.) Startposition in Pfad-Array uebernehmen
        NSMutableArray *singleElectronPath = [[[NSMutableArray alloc] init]
                                              autorelease];
        [singleElectronPath addObject:[NSValue valueWithCGPoint:
                                       CGPointMake(singleElectron.position.x+1,
                                                   singleElectron.position.y+1)]];
        // 2.) Zielposition in Pfad-Array uebernehmen
        // Zufallsposition irgendwo an oberen oder unteren Rand
        int x = fmod(arc4random(),screenWidth);
        int y = (arc4random()%2?0:screenHeight);
        [singleElectronPath addObject:[NSValue valueWithCGPoint:CGPointMake(x, y)]];
        [electronPath addObject:singleElectronPath];
    }
    
    // Zoom auf 100% setzen
    self.scale=1.0;
    
    // frame Zaehler
    counter = 0;
    
    // currentElektron = aktuell zu steuerndes Elektron
    // (-1 = kein Elektron ausgewaehlt)
    currentElectron = -1;
    
    // hier wird ID des Elektrons gespeichert,
    // welches das Spiel beendet hat und auf
    // das hingezoomt wird (Ausnahme: thereIsNoBadElectron = Sieg)
    zoomToBadElectron = -1;
    
    // Gamemodus aktivieren
    normalAction = YES;
}

-(id) init
{
    // Anzahl der Elektronen fuer Level 1
    electronsTotal = 4;
    
    if( (self=[super init])) {
        
        // Display beruehrungsempfindlich setzen
        self.isTouchEnabled = YES;
        
        // FPS-Anzeige abschalten
        CCDirectorIOS* director = (CCDirectorIOS*)[CCDirector sharedDirector];
        director.displayStats = NO;
        
        // Displaybreite und -hoehe ermitteln
        CGSize size = [director winSize];
        screenHeight = size.height;
        screenWidth = size.width;
        
        // Alle Array intialisieren
        bg = [[NSMutableArray alloc] init];
        atoms = [[NSMutableArray alloc] init];
        electrons = [[NSMutableArray alloc] init];
        electronPath = [[NSMutableArray alloc] init];
        
        // Leveldesign setzen
        [self restartLevel];
        
        // GameEngine starten
        [self schedule:@selector(gameLoop:) interval: 1/60.0f];
        
    }
    return self;
}

// Diese Methode erspart uns etwas Schreiberei
// fuer Punkt-Punkt-Abstandsberechnung (z. B. fuer Kollisionen)
- (bool) isDistanceFrom: (CGPoint) x to:(CGPoint) y lessThan:(float) d {
    if ((x.x-y.x)*(x.x-y.x)+(x.y-y.y)*(x.y-y.y)<d*d) return YES; else return NO;
}


-(void) animateBackground {
    for (int currentId=0; currentId<[bg count]; currentId++) {
        // Sprite
        CCSprite *bgAtom = [bg objectAtIndex:currentId];
        
        // Atom je nach Index verkleinern
        bgAtom.scale = 0.025*(currentId+1);
        
        // Atom wie Luftblasen sinusförmig nach oben bewegen
        bgAtom.position = ccp(bgAtom.position.x+cos(currentId+bgAtom.position.y/20),
                              bgAtom.position.y+0.5);
        // Wenn Atom oben ankommt, unten neu setzen
        if (bgAtom.position.y > screenHeight+150) bgAtom.position =
            ccp(fmod(arc4random(),screenWidth), -150);
        
        // Transparenz kosinusartig verändern
        bgAtom.opacity = 60.0+cos(counter*(currentId+1))*20;
    }
}

// bewegt die Atome
-(void) moveAtoms {
    
    // beide Atome
    for (int atomId=0;atomId<2;atomId++) {
        
        // darin jeweils alle 3 Sprites (ueberlagert)
        for (int atomLayer=0; atomLayer<3; atomLayer++) {
            float counterTmp = counter*3+atomId;
            CCSprite *singleAtom = [[atoms objectAtIndex:atomId]
                                    objectAtIndex:atomLayer];
            
            // minimal aufeinander zu bewegen
            if (100+counter*200<screenWidth/2) {
                singleAtom.position =  ccp( (atomId==0?100+counter*200:
                                             (screenWidth-100)-counter*200),
                                           screenHeight/2 );
            }
            
            // oder auf Endposition: gewonnen!
            else if (zoomToBadElectron==-1) {
                normalAction=NO;
                // 2 Elektronen mehr im neuen Level
                electronsTotal+=2;
                //
                zoomToBadElectron=thereIsNoBadElectron;
            }
            
            // unterschiedliche Drehgrade
            float angleRadians = atomLayer+counterTmp*(1-atomLayer+0.5);
            float angleDegrees = CC_RADIANS_TO_DEGREES(angleRadians);
            singleAtom.rotation=angleDegrees;
            
            // Halbtransparenz pulsieren (Werte: 100...200)
            singleAtom.opacity = 150.0+cos(10*counterTmp*(atomLayer+1))*50;
        }
    }
}


-(void) moveElectrons {
    
    // nur bewegen, falls Gamemode aktiv
    if (normalAction) {
        
        // alle Elektronen durchlaufen
        for (int currentId=0;currentId<[electrons count];currentId++) {
            
            // Elektron holen
            CCSprite *singleElectron = [electrons objectAtIndex:currentId];
            
            // erste Zielposition im Pfad-Array holen
            CGPoint goToPoint = [[[electronPath objectAtIndex:currentId]
                                  objectAtIndex:0] CGPointValue];
            
            // Wenn erste Zielposition erreicht
            if([self isDistanceFrom:singleElectron.position
                                 to:goToPoint
                           lessThan:1.0])
            {
                // Zielpunkt entfernen, die dahinter liegenden
                // ruecken eine Position auf
                if ([[electronPath objectAtIndex:currentId] count] > 1) {
                    [[electronPath objectAtIndex:currentId] removeObjectAtIndex:0];
                }
                goToPoint = [[[electronPath objectAtIndex:currentId] objectAtIndex:0]
                             CGPointValue];
            }
            
            // abgewandelter Bresenham-Algorithmus:
            // gleichmaessig zu naechstem Zielpunkt bewegen
            float deltaX = goToPoint.x - singleElectron.position.x;
            float deltaY = goToPoint.y - singleElectron.position.y;
            
            float dx=1.0/decElectronSpeed;
            float dy=1.0/decElectronSpeed;
            
            if (deltaX==0) {dx=0;}
            else if (deltaY==0) {dy=0;}
            else if (fabs(deltaX) > fabs(deltaY)) {
                dy=fabs(deltaY/deltaX)/decElectronSpeed;
            }
            else if (fabs(deltaX) < fabs(deltaY)) {
                dx=fabs(deltaX/deltaY)/decElectronSpeed;
            }
            
            if (deltaX<0) {dx=-dx;}
            if (deltaY<0) {dy=-dy;}
            
            // Elektron auf neue Position setzen
            singleElectron.position = ccp(singleElectron.position.x + dx,
                                          singleElectron.position.y + dy);
            
        }
        
    }
}

// Elektronen auf Verfehlungen pruefen
-(void) checkIfNukeIsUnderControl {
    
    // Nur im Gamemodus aktiv werden
    if (normalAction) {
        
        // alle Elektronen durchchecken
        for (int currentId=0; currentId<[electrons count]; currentId++) {
            
            // Elektron holen
            CCSprite *singleElectron = [electrons objectAtIndex:currentId];
            
            // Fehler 1:
            // auf Kollision mit anderen Elektronen pruefen
            for (int i=currentId+1;i<[electrons count];i++) {
                CCSprite *t1 = [electrons objectAtIndex:i];
                if ([self isDistanceFrom:t1.position
                                      to:singleElectron.position
                                lessThan:13.0]) {
                    normalAction=NO;
                    
                    // Wenn getroffen: Zoom zum Uebeltaeter
                    if (zoomToBadElectron==-1) zoomToBadElectron=currentId;
                }
            }
            
            // Fehler 2:
            // Elektron  noch innerhalb beider Atome?
            CCSprite *atom1 = [[atoms objectAtIndex:0] objectAtIndex:0];
            CCSprite *atom2 = [[atoms objectAtIndex:1] objectAtIndex:0];
            if (
                (![self isDistanceFrom:atom1.position
                                    to:singleElectron.position
                              lessThan:135.0]) &&
                (![self isDistanceFrom:atom2.position
                                    to:singleElectron.position
                              lessThan:135.0])
                ) {
                normalAction=NO;
                
                // Wenn getroffen: Zoom zum Uebeltaeter
                if (zoomToBadElectron==-1) zoomToBadElectron=currentId;
            }
        }
    }
}

// Zoom-Modus
-(void) andDoThisIfNukeIsOutOfControl {
    
    // nur Aktiv, wenn Level gewonnen/verloren
    if (zoomToBadElectron>-1) {
        
        // Hauptlayer reinzoomen
        self.scale=self.scale*1.05;
        
        // Wenn gewonnen: auf Displaymitte!
        CGPoint zoomToPoint = ccp(0.5,0.5);
        
        // Wenn verloren: auf Uebeltaeter-Elektron
        if (zoomToBadElectron!=thereIsNoBadElectron) {
            CCSprite *badElectron = [electrons objectAtIndex:zoomToBadElectron];
            zoomToPoint = ccp(badElectron.position.x/screenWidth,
                              badElectron.position.y/screenHeight);
        }
        
        // Zoommittelpunkt setzen
        self.anchorPoint=zoomToPoint;
        
        // Nur bis Vergroesserungsfaktor 5 zoomen
        // dann Levelneustart
        if (self.scale>5) [self restartLevel];
    }
}

// Game-Engine
- (void) gameLoop: (ccTime) dT {
    counter=counter+0.0005;
    [self animateBackground];
    [self moveAtoms];
    [self moveElectrons];
    [self checkIfNukeIsUnderControl];
    [self andDoThisIfNukeIsOutOfControl];
}

// Zeichenfunktionen fuer aktuellen Layer
-(void) draw {
    
    // Liniendicke definieren: 2 Punkte
    glLineWidth(2.0f);
    
    // alle Pfadpunkte durchlaufen
    for (int currentId=0; currentId<[electrons count]; currentId++) {
        CCSprite *singleElectron = [electrons objectAtIndex:currentId];
        
        // Pfad zeichnen
        if ([[electronPath objectAtIndex:currentId] count]>2) {
            
            // von Elektronposition ausgehend
            ccDrawLine(CGPointMake(singleElectron.position.x,
                                   singleElectron.position.y),
                       [[[electronPath objectAtIndex:currentId] objectAtIndex:0] CGPointValue]);
            
            // alle Punkte durchlaufen und miteinander verbinden
            for (int singlePathPoint = 0;
                 singlePathPoint < [[electronPath objectAtIndex:currentId] count] - 2;
                 singlePathPoint++) {
                CGPoint start=[[[electronPath objectAtIndex:currentId]
                                objectAtIndex:singlePathPoint] CGPointValue];
                CGPoint end=[[[electronPath objectAtIndex:currentId]
                              objectAtIndex:singlePathPoint+1] CGPointValue];
                // Linie zeichnen
                ccDrawLine(start,end);
            }
        }
        
    }
    
}

// TouchesBegan fuer Cocos2d
-(void) ccTouchesBegan:(NSSet *)touches withEvent:(UIEvent *)event {
    if (normalAction) {
        
        // Beruehrpunkt holen
        UITouch *myTouch = [touches anyObject];
        CGPoint point = [myTouch locationInView:[myTouch view]];
        point = [[CCDirector sharedDirector] convertToGL:point];
        
        // alle Elektronen durchlaufen
        for (int currentId=0; currentId<[electrons count]; currentId++) {
            CCSprite *singleElectron = [electrons objectAtIndex:currentId];
            
            // Wenn Beruehrpunkt in Naehe eines Elektrons:
            if ([self isDistanceFrom:singleElectron.position to:point lessThan:25.0]){
                
                // dessen Id auf current setzen
                currentElectron=currentId;
                
                // bisherigen Pfad loeschen
                nodeCounter=0;
                [[electronPath objectAtIndex:currentElectron] removeAllObjects];
                
                // ersten Punkt des Pfads setzen (= Koordinate des Elektrons)
                // (damit bei ungenauem ersten Beruehrpunkt Elektron nicht
                // erstmal eine Kante abfaehrt oder sich in Gegenrichtung bewegt)
                [[electronPath objectAtIndex:currentElectron]
                 addObject:[NSValue valueWithCGPoint:
                            CGPointMake(singleElectron.position.x, singleElectron.position.y)]];
            }
        }
    }
}

// TouchesMoved fuer Cocos2d
-(void) ccTouchesMoved:(NSSet *)touches withEvent:(UIEvent *)event {
    if ((normalAction)&&(currentElectron>-1)) {
        
        // Beruehrpunkt holen
        UITouch *myTouch = [touches anyObject];
        CGPoint point = [myTouch locationInView:[myTouch view]];
        point = [[CCDirector sharedDirector] convertToGL:point];
        
        // aktuell aktives Elektron holen
        CCSprite *singleElectron = [electrons objectAtIndex:currentElectron];
        
        // Pfadpunkte nur setzen, wenn Koordinaten ausserhalb Elektrons
        // (damit bei ungenauem ersten Beruehrpunkt Elektron nicht
        // erstmal eine Kante abfaehrt oder sich in Gegenrichtung bewegt)
        if ((currentElectron > -1) &&
            (![self isDistanceFrom:singleElectron.position to:point lessThan:25.0])){
            nodeCounter++;
            
            // nur jeden dritten Punkt aufnehmen (weniger Speicher)
            if (nodeCounter%2==0) [[electronPath objectAtIndex:currentElectron]
                                   addObject:[NSValue valueWithCGPoint:CGPointMake(point.x, point.y)]];
            
        }
    }
}

// TouchesEnded fuer Cocos2d
-(void) ccTouchesEnded:(NSSet *)touches withEvent:(UIEvent *)event {
    if ((normalAction)&&(currentElectron>-1)) {
        
        // Beruehrpunkt holen
        UITouch *myTouch = [touches anyObject];
        CGPoint point = [myTouch locationInView:[myTouch view]];
        point = [[CCDirector sharedDirector] convertToGL:point];
        
        // verlaengert die letzten beiden Pfadpunkte in "Unendlichkeit"
        // damit sich Elektron in zuletzt eingeschlagene Richtung
        // weiter bewegt
        
        // letzten Pfadpunkt holen
        CGPoint end = [[[electronPath objectAtIndex:currentElectron] lastObject]
                       CGPointValue];
        
        // wenn Pfad zu kurz (nur ein gespeicherter Punkt),
        // dann aktuellen Beruehrpunkt hinzufuegen
        if ([[electronPath objectAtIndex:currentElectron] count]<2) {
            [[electronPath objectAtIndex:currentElectron]
             addObject:[NSValue valueWithCGPoint:CGPointMake(point.x, point.y)]];
        }
        
        // vorletzten Pfadpunkt holen
        CGPoint start = [[[electronPath objectAtIndex:currentElectron]
                          objectAtIndex:[[electronPath
                                          objectAtIndex:currentElectron] count]-2]
                         CGPointValue];
        
        // Punkt in Verlaengerung ausserhalb des Viewports berechnen
        float x=start.x+screenWidth*(end.x-start.x);
        float y=start.y+screenWidth*(end.y-start.y);
        
        // berechneten Punkt an Pfad anfuegen
        [[electronPath objectAtIndex:currentElectron] addObject:
         [NSValue valueWithCGPoint:CGPointMake(x,y)]];
        
        // aktuelles Elektron auf inaktiv setzen
        currentElectron = -1;
    }
}

- (void) dealloc {
    [super dealloc];
    [atoms release];
    [electronPath release];
    [electrons release];
    [bg release];
}
@end
